#include <windows.h>
#include "cube.h"

// decode functions (the stuff that I had the least to do with)

// (from in_dsp)
//================ C R A Z Y N A T I O N ===================================
//==========================================================================
// ADPCM decoder for Nintendo GAMECUBE dsp format, w/ modifications by hcs
//==========================================================================
// decode 8 bytes of input (1 frame, 14 samples)
int DSPdecodebuffer
(
    u8			*input, // location of encoded source samples
    s16         *out,   // location of destination buffer (16 bits / sample)
    short		coef[16],   // location of adpcm info
	short * histp,
	short * hist2p
)
{
int sample;
short nibbles[14];
int index1;
int i,j;
char *src,*dst;
short delta;
short hist=*histp;
short hist2=*hist2p;

	dst = (char*)out;

	src=input;
    i = *src&0xFF;
    src++;
    delta = 1 << (i & 255 & 15);
    index1 = (i & 255) >> 4;

    for(i = 0; i < 14; i = i + 2) {
		j = ( *src & 255) >> 4;
		nibbles[i] = j;
		j = *src & 255 & 15;
		nibbles[i+1] = j;
		src++;
	}

     for(i = 0; i < 14; i = i + 1) {
		if(nibbles[i] >= 8) 
			nibbles[i] = nibbles[i] - 16;       
     }
     
	 for(i = 0; i<14 ; i = i + 1) {

		sample = (delta * nibbles[i])<<11;
		sample += coef[index1*2] * hist;
		sample += coef[index1*2+1] * hist2;
		sample = sample + 1024;
		sample = sample >> 11;
		if(sample > 32767) {
			sample = 32767;
		}
		if(sample < -32768) {
			sample = -32768;
		}

        *(short*)dst = (short)sample;
        dst = dst + 2;

		hist2 = hist;
        hist = (short)sample;
        
    }
	*histp=hist;
	*hist2p=hist2;

    return((int)src);
}

#define BASEVOL 0x2000

// (from in_adx)
/************** Elite ANSI C ADX decoder function **************/
// decode 18 bytes of input (32 frames)
int ADXdecodebuffer(unsigned char *in, short *out, short *hist1, short *hist2)
{

  int scale = ( (in[0] << 8) | (in[1]) );
  int i;
  int s0, s1, s2, d;


  in += 2;


  s1 = *hist1;
  s2 = *hist2;

  for (i = 0; i < 16; i++)
  {
    d = in[i] >> 4;

    if (d & 8)
    {
      d -= 16;
    }

    s0 = (BASEVOL*d*scale + 0x7298*s1 - 0x3350*s2) >> 14;

    if (s0 > 32767)
    {
      s0 = 32767;
    }
    else if (s0 < -32768)
    {
      s0 = -32768;
    }

	*out++ = s0;
  
    s2 = s1;
    s1 = s0;

    d = in[i] & 15;

    if (d & 8)
    {
      d -= 16;
    }

    s0 = (BASEVOL*d*scale + 0x7298*s1 - 0x3350*s2) >> 14;

    if (s0 > 32767)
    {
      s0 = 32767;
    }
    else if (s0 < -32768)
    {
      s0 = -32768;
    }

  	*out++ = s0;
	
    s2 = s1;
    s1 = s0;

  }

  *hist1 = s1;
  *hist2 = s2;

  return 0;
   
}

// (from gc2wav/cube2wav)
/*********************************************************************

        Nintendo GameCube ADPCM Decoder
        Author: Shinji Chiba <ch3@mail.goo.ne.jp>

*********************************************************************/

#define ONE_BLOCK_SIZE		32
#define SAMPLES_PER_BLOCK	28

short ADPDecodeSample( int bits, int q, float * hist1, float * hist2 )
{
	static const float coef[4] = { 0.86f, 1.8f, 0.82f, 1.53f };
	float iir_filter;

	iir_filter = (float) ((short) (bits << 12) >> (q & 0xf));
	switch( q >> 4 )
	{
	case 1:
		iir_filter += (*hist1 * coef[0]);
		break;
	case 2:
		iir_filter += (*hist1 * coef[1] - *hist2 * coef[2]);
		break;
	case 3:
		iir_filter += (*hist1 * coef[3] - *hist2 * coef[0]);
	}

	*hist2 = *hist1;
	if ( iir_filter <= -32768.5f )
	{
		*hist1 = -32768.5f;
		return -32768;
	}
	else if ( iir_filter >= 32767.5f )
	{
		*hist1 = 32767.5f;
		return 32767;
	}
	return (short) (*hist1 = iir_filter);
}

// decode 32 bytes of input (28 samples), assume stereo
int ADPdecodebuffer(unsigned char *input, short *outl, short * outr, float *histl1, float *histl2, float *histr1, float *histr2) {
	int i;
	for( i = 0; i < SAMPLES_PER_BLOCK; i++ )
	{
		outl[i] = ADPDecodeSample( input[i + (ONE_BLOCK_SIZE - SAMPLES_PER_BLOCK)] & 0xf, input[0], histl1, histl2 );
		outr[i] = ADPDecodeSample( input[i + (ONE_BLOCK_SIZE - SAMPLES_PER_BLOCK)] >> 4, input[1], histr1, histr2 );
	}
	return 0;
}
